import base64
from typing import Iterable, Union

STANDARD_ALPHABET = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ234567'
CUSTOM_ALPHABET = 'abcdefghjkmnprstuvwxyz0123456789'
ENCODE_TRANS = str.maketrans(STANDARD_ALPHABET, CUSTOM_ALPHABET)
DECODE_TRANS = str.maketrans(CUSTOM_ALPHABET, STANDARD_ALPHABET)
PADDING_LETTER = '='


def encode(buffer: Union[bytes, bytearray]) -> str:
    assert type(buffer) == bytes or type(buffer) == bytearray, "please pass an bytes"
    b32encoded = base64.b32encode(buffer)  # encode bytes
    b32str = b32encoded.decode().replace(PADDING_LETTER, "")  # translate chars
    return b32str.translate(ENCODE_TRANS)  # remove padding char


def decode(b32str: str) -> bytes:
    if not isinstance(b32str, str):
        raise TypeError(f"Invalid argument type: base32 decode requires a string type argument "
                        f"but receives an argument of type {type(b32str)}")
    if b32str != b32str.lower():
        raise ValueError(f"Invalid value: only lower case letters are used for base32 address, "
                        f"receives {b32str}")
    # pad to 8's multiple with '='
    b32len = len(b32str)
    if b32len % 8 > 0:
        padded_len = b32len + (8 - b32len % 8)
        b32str = b32str.ljust(padded_len, PADDING_LETTER)
    # translate and decode
    return base64.b32decode(b32str.translate(DECODE_TRANS))


def decode_to_words(b32str: str) -> bytearray:
    result = bytearray()
    for c in b32str:
        result.append(CUSTOM_ALPHABET.index(c))
    return result


def encode_words(words: Iterable[int]) -> str:
    result = ""
    for v in words:
        result += CUSTOM_ALPHABET[v]
    return result



from typing import (
    Any
)

from cfx_address._utils import (
    validate_hex_address,
    validate_network_id,
    eth_eoa_address_to_cfx_hex
)
from cfx_address.address import (
    Base32Address
)
# from eth_utils.address import (
#     is_hex_address
# )

validate_base32 = Base32Address.validate
is_valid_base32 = Base32Address.is_valid_base32

# def is_valid_address(value: str) -> bool:
#     """
#     checks if a value is a valid string-typed address, either hex address or base32 address is ok

#     :param Any value: value to check
#     :return bool: True if valid, otherwise False
#     """    
#     return is_hex_address(value) or is_valid_base32(value)

__all__ = [
    "validate_hex_address",
    "validate_network_id",
    "eth_eoa_address_to_cfx_hex",
    "validate_base32",
    "is_valid_base32",
    # "is_hex_address"
]



# utils do not rely on other modules are defined here to avoid recursive import
from typing import Any, Literal

from cfx_utils.exceptions import (
    InvalidNetworkId,
    InvalidHexAddress, 
    
)
from eth_utils.address import (
    is_hex_address
)
from eth_typing.evm import (
    HexAddress,
)

def eth_eoa_address_to_cfx_hex(eoa_address: str) -> HexAddress:
    """
    Convert an ethereum EOA address to valid cfx hex address.
    
    In conflux, only addresses starting with 0x1 are valid user-type addresses. 
    This function convert ethereum EOA address to the corresponding form in conflux.

    :param str address: ethereum address
    :raises InvalidHexAddress: the argument is not a valid hex address 
    :return HexAddress: corresponding hex address in conflux, starting with '0x1'
    :examplse: 
    
    >>> eth_eoa_address_to_cfx_hex("0xd43d2a93e97245E290feE74276a1EF8D275bE646")
    '0x143d2a93e97245e290fee74276a1ef8d275be646'
    """
    validate_hex_address(eoa_address)
    return '0x1' + eoa_address.lower()[3:] # type: ignore


def hex_address_bytes(hex_address: str) -> bytes:
    assert type(hex_address) == str
    return bytes.fromhex(hex_address.lower().replace('0x', ""))

def validate_network_id(network_id: Any) -> Literal[True]:
    """
    Checks if the given value is a positive integer.


    :param Any hex_address: value to validate
    :raises InvalidHexAddress: raised if not valid
    :return Literal[True]: returns True if valid
    """
    if isinstance(network_id, int) and network_id > 0:
        return True
    raise InvalidNetworkId("Expected network_id to be a positive integer. "
                     f"Receives {network_id} of type {type(network_id)}")

def validate_hex_address(value: Any) -> Literal[True]:
    """
    Checks if the given string of text type is an address in hexadecimal encoded form.

    :param Any hex_address: value to validate
    :raises InvalidHexAddress: raised if not valid
    :return Literal[True]: returns True if valid
    """
    if not is_hex_address(value):
        raise InvalidHexAddress("Expected a hex40 address. "
                                f"Receives {value}")
    return True



MAINNET_NETWORK_ID = 1029
TESTNET_NETWORK_ID = 1
TYPE = "type"
MAINNET_PREFIX = "cfx"
TESTNET_PREFIX = "cfxtest"
COMMON_NET_PREFIX = "net"

TYPE_NULL = "null"
TYPE_BUILTIN = "builtin"
TYPE_USER = "user"
TYPE_CONTRACT = "contract"
TYPE_INVALID = "invalid"

HEX_PREFIX = "0x"
DELIMITER = ":"

VERSION_BYTE = bytes(1)
CHECKSUM_TEMPLATE = bytes(8)



import sys

from typing import (
    Literal,
    NewType, 
    TypedDict,
    Union
)

from eth_typing import HexAddress

if sys.version_info >= (3,9):
    from typing import Annotated
    TRIVIAL_NETWORK_PREFIX = Annotated[str, lambda x: x.startswith("net")]
else:
    TRIVIAL_NETWORK_PREFIX = NewType("TRIVIAL_NETWORK_PREFIX", str)

NetworkPrefix = Union[
    Literal["cfx", "cfxtest"], TRIVIAL_NETWORK_PREFIX,
]

AddressType = Literal[
    "null", 
    "builtin", 
    "user", 
    "contract", 
    "invalid"
]

class Base32AddressParts(TypedDict):
    network_id: int
    address_type: AddressType
    hex_address: HexAddress



from setuptools import setup, find_packages

DESCRIPTION = 'Conflux base32 address'


extras_require = {
    'tester': [
        "pytest>=6.2.5,<7",
    ],
    'linter': [
        # "black>=22.1.0,<23.0",
        # "flake8==3.8.3",
        # "isort>=4.2.15,<4.3.5",
        # "mypy==0.910",
        # "types-setuptools>=57.4.4,<58",
        # "types-requests>=2.26.1,<3",
        # "types-protobuf==3.19.13",
    ],
    'docs': [
        # "mock",
        # "sphinx-better-theme>=0.1.4",
        # "click>=5.1",
        # "configparser==3.5.0",
        # "contextlib2>=0.5.4",
        # "py-geth>=3.8.0,<4",
        # "py-solc>=0.4.0",
        # "pytest>=6.2.5,<7",
        "sphinx>=4.2.0,<5",
        "sphinx_rtd_theme>=0.1.9",
        # "toposort>=1.4",
        # "towncrier==18.5.0",
        # "urllib3",
        # "wheel"
    ],
    'dev': [
        # "bumpversion",
        "wheel",
        "pycryptodome",
        "bumpversion==0.6.0"
        # "flaky>=3.7.0,<4",
        # "hypothesis>=3.31.2,<6",
        # "pytest>=6.2.5,<7",
        # "pytest-asyncio>=0.18.1,<0.19",
        # "pytest-mock>=1.10,<2",
        # "pytest-pythonpath>=0.3",
        # "pytest-watch>=4.2,<5",
        # "pytest-xdist>=1.29,<2",
        # "setuptools>=38.6.0",
        # "tox>=1.8.0",
        # "tqdm>4.32,<5",
        # "twine>=1.13,<2",
        # "pluggy==0.13.1",
        # "when-changed>=0.3.0,<0.4"
    ]
}

extras_require['dev'] = (
    extras_require['tester']
    + extras_require['linter']
    + extras_require['docs']
    + extras_require['dev']
)

with open('./README.md') as readme:
    long_description = readme.read()

# Setting up
setup(
    # the name must match the folder name 'verysimplemodule'
    name="cfx-address",
    version="1.0.0-beta.4", # edit using bumpversion
    author="The Conflux foundation",
    author_email="wangpan@conflux-chain.org",
    long_description_content_type='text/markdown',
    long_description=long_description,
    packages=find_packages(),
    install_requires=[
        "eth-utils>=2.0.0,<3.0.0",
        "hexbytes>=0.1.0,<1.0.0",
        "cfx-utils>=1.0.0b3",
    ],  # add any additional packages that
    # needs to be installed along with your package. Eg: 'caer'
    extras_require=extras_require,
    keywords=['Conflux', 'base32', 'address'],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Education",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)



from typing import (
    Any,
    Literal,
    Optional, 
    Union,
    cast,
    get_args
)

from hexbytes import HexBytes
from eth_typing.evm import (
    ChecksumAddress,
    HexAddress,
)
from eth_utils.crypto import (
    keccak
)
from eth_utils.address import (
    to_checksum_address,
    is_hex_address
)

from cfx_address import (
    base32
)
from cfx_address._utils import (
    hex_address_bytes,
    validate_hex_address,
    validate_network_id
)
from cfx_address.types import (
    AddressType,
    Base32AddressParts,
    NetworkPrefix,
)
from cfx_utils.exceptions import (
    InvalidAddress,
    InvalidBase32Address,
    InvalidConfluxHexAddress,
)
from cfx_address.consts import (
    MAINNET_NETWORK_ID,
    TESTNET_NETWORK_ID,
    TYPE,
    MAINNET_PREFIX ,
    TESTNET_PREFIX,
    COMMON_NET_PREFIX,

    TYPE_NULL,
    TYPE_BUILTIN,
    TYPE_USER,
    TYPE_CONTRACT,
    TYPE_INVALID,
    HEX_PREFIX,
    DELIMITER,
    
    VERSION_BYTE,
    CHECKSUM_TEMPLATE,
)


class Base32Address(str):
    """
    Class Base32Address can be used to create Base32Address instances and provides useful class methods to deal with base32 format addresses.
    ## Base32Address inherits from str, so the Base32Address can be trivially used as strings
    
    :examples:
    
    >>> address = Base32Address("0x1ecde7223747601823f7535d7968ba98b4881e09", network_id=1)
    >>> address
    'cfxtest:aatp533cg7d0agbd87kz48nj1mpnkca8be1rz695j4'
    >>> [
    ...     address.address_type,
    ...     address.network_id,
    ...     address.hex_address,
    ...     address.verbose_address,
    ...     address.abbr,
    ...     address.mapped_evm_space_address,
    ...     address.eth_checksum_address,
    ... ]
    ['user', 1, '0x1ecde7223747601823f7535d7968ba98b4881e09', 'CFXTEST:TYPE.USER:AATP533CG7D0AGBD87KZ48NJ1MPNKCA8BE1RZ695J4', 'cfxtest:aat...95j4', '0x349f086998cF4a0C5a00b853a0E93239D81A97f6', '0x1ECdE7223747601823f7535d7968Ba98b4881E09']
    """    
    
    def __new__(cls, address: Union["Base32Address", HexAddress, str], network_id: Optional[int]=None, verbose: bool = False) -> "Base32Address":
        """
        :param Union[Base32Address, HexAddress, str] address: a base32-or-hex format address
        :param Optional[int] network_id: target address network_id. Optional if first argument is a base32 address, but required for hex address
        :param bool verbose: whether the return value will be encoded in verbose mode, defaults to False
        :raises InvalidAddress: address is neither base32 address nor hex address
        :raises InvalidNetworkId: network_id argument is not a positive integer or is None when address argument is a hex address
        :return Base32Address: an encoded base32 object, which can be trivially used as python str
        
        :examples:
        
        >>> address = Base32Address("0x1ecde7223747601823f7535d7968ba98b4881e09", network_id=1)
        >>> address
        'cfxtest:aatp533cg7d0agbd87kz48nj1mpnkca8be1rz695j4'
        >>> address_ = Base32Address(address, network_id=1029, verbose=True)
        >>> address_
        'CFX:TYPE.USER:AATP533CG7D0AGBD87KZ48NJ1MPNKCA8BE7GGP3VPU'
        >>> isinstance(address_, str)
        True
        """
        try:
            parts = cls.decode(address)
            if network_id is None:
                val = cls._encode(parts["hex_address"], parts["network_id"], verbose)
            else:
                validate_network_id(network_id)    
                val = cls._encode(parts["hex_address"], network_id, verbose)
        except InvalidBase32Address:
            if is_hex_address(address):
                validate_network_id(network_id)
                val = cls._encode(address, network_id, verbose) # type: ignore
            else:
                raise InvalidAddress("Address should be either base32 or hex, "
                                f"receives {address}")
        
        return str.__new__(cls, val)
    
    def __eq__(self, _address: str) -> bool:
        """
        invoked when a base32 address is compared to another string (or Base32Address typed object),
        
        :param str __address: value compared to 
        :return bool: True if self and _address are of same hex_address and network_id
        :raises InvalidBase32Address: _address is not encoded in base32 format
        :examples:
        
        >>> address = Base32Address("cfxtest:aatp533cg7d0agbd87kz48nj1mpnkca8be1rz695j4")
        >>> assert address == "CFXTEST:TYPE.USER:AATP533CG7D0AGBD87KZ48NJ1MPNKCA8BE1RZ695J4"
        >>> assert "CFXTEST:TYPE.USER:AATP533CG7D0AGBD87KZ48NJ1MPNKCA8BE1RZ695J4" == address
        >>> assert "cfxtest:aatp533cg7d0agbd87kz48nj1mpnkca8be1rz695j4" == address
        """
        parts = Base32Address.decode(_address)
        return self.hex_address == parts["hex_address"] and self.network_id == parts["network_id"]

    def __hash__(self) -> int:
        return super().__hash__()

    @property
    def network_id(self) -> int:
        """
        :return int: network_id of the address
        :examples:
        
        >>> address = Base32Address("cfxtest:aatp533cg7d0agbd87kz48nj1mpnkca8be1rz695j4")
        >>> address.network_id
        1
        """        
        return Base32Address._decode_network_id(self)

    @property
    def hex_address(self) -> HexAddress:
        """
        :return HexAddress: hex address of the address.
        
        >>> address = Base32Address("cfxtest:aatp533cg7d0agbd87kz48nj1mpnkca8be1rz695j4")
        >>> address.hex_address
        '0x1ecde7223747601823f7535d7968ba98b4881e09'
        """        
        return Base32Address._decode_hex_address(self)

    @property
    def address_type(self) -> AddressType:
        """
        :return AddressType: address type of an address
        
        :examples:
        
        >>> address = Base32Address("cfx:aatp533cg7d0agbd87kz48nj1mpnkca8be7ggp3vpu")
        >>> address.address_type
        'user'
        """      
        return Base32Address._detect_address_type(hex_address_bytes(self.hex_address))

    @property
    def eth_checksum_address(self) -> ChecksumAddress:
        """
        :return ChecksumAddress: self.hex_address in ethereum checksum address format
        
        >>> address = Base32Address("cfxtest:aatp533cg7d0agbd87kz48nj1mpnkca8be1rz695j4")
        >>> address.hex_address
        '0x1ecde7223747601823f7535d7968ba98b4881e09'
        >>> address.eth_checksum_address
        '0x1ECdE7223747601823f7535d7968Ba98b4881E09'
        """        
        return to_checksum_address(self.hex_address)

    @property
    def verbose_address(self) -> "Base32Address":
        """
        :return Base32Address: self presented in verbose mode
        
        >>> address = Base32Address("cfxtest:aatp533cg7d0agbd87kz48nj1mpnkca8be1rz695j4")
        >>> address.verbose_address
        'CFXTEST:TYPE.USER:AATP533CG7D0AGBD87KZ48NJ1MPNKCA8BE1RZ695J4'
        """        
        return Base32Address.encode(self.hex_address, self.network_id, True)
    
    @property
    def abbr(self) -> str:
        """
        :return str: abbreviation of the address, as mentioned in https://forum.conflux.fun/t/voting-results-for-new-address-abbreviation-standard/7131
        
        :examples:
        
        >>> Base32Address("cfxtest:aatp533cg7d0agbd87kz48nj1mpnkca8be1rz695j4").abbr
        'cfxtest:aat...95j4'
        >>> Base32Address("cfx:aatp533cg7d0agbd87kz48nj1mpnkca8be7ggp3vpu").abbr
        'cfx:aat...7ggp3vpu'
        """        
        return Base32Address._shorten_base32_address(self)
    
    @property
    def compressed_abbr(self) -> str:
        """
        :return str: compressed abbreviation of the address, same as property "abbr" except for mainnet address
             
        :examples:
        
        >>> address = Base32Address("cfx:aatp533cg7d0agbd87kz48nj1mpnkca8be7ggp3vpu")
        >>> address.abbr
        'cfx:aat...7ggp3vpu'
        >>> address.compressed_abbr
        'cfx:aat...3vpu'
        """ 
        return Base32Address._shorten_base32_address(self, True)
    
    @property
    def mapped_evm_space_address(self) -> HexAddress:
        """
        :return HexAddress: the address of mapped account for EVM space as defined in https://github.com/Conflux-Chain/CIPs/blob/master/CIPs/cip-90.md#mapped-account
        
        >>> address = Base32Address("cfx:aatp533cg7d0agbd87kz48nj1mpnkca8be7ggp3vpu")
        >>> address.mapped_evm_space_address
        '0x349f086998cF4a0C5a00b853a0E93239D81A97f6'
        """
        return Base32Address._mapped_evm_address_from_hex(self.hex_address)

    @classmethod
    def encode(
        cls, hex_address: str, network_id: int, verbose: bool=False
    ) -> "Base32Address":
        """
        encode hex address to base32 address

        :param str hex_address: hex address begins with 0x
        :param int network_id: address network id, e.g., 1 for testnet and 1029 for mainnet
        :param bool verbose: whether the address will be presented in verbose mode, defaults to False
        :return Base32Address: an encoded base32 object, which can be trivially used as python str
        
        :examples:
        
        >>> Base32Address.encode("0x1ecde7223747601823f7535d7968ba98b4881e09", 1)
        'cfxtest:aatp533cg7d0agbd87kz48nj1mpnkca8be1rz695j4'
        >>> Base32Address.encode("0x1ecde7223747601823f7535d7968ba98b4881e09", 1029, verbose=True)
        'CFX:TYPE.USER:AATP533CG7D0AGBD87KZ48NJ1MPNKCA8BE7GGP3VPU'
        """ 
        validate_hex_address(hex_address)
        validate_network_id(network_id)
        return Base32Address(hex_address, network_id, verbose)

    @classmethod
    def _encode(
        cls, hex_address: str, network_id: int, verbose: bool=False
    ) -> str:
        network_prefix = cls._encode_network_prefix(network_id)
        address_bytes = hex_address_bytes(hex_address)
        payload = base32.encode(VERSION_BYTE + address_bytes)
        checksum = cls._create_checksum(network_prefix, payload)
        parts = [network_prefix]
        address_type = cls._detect_address_type(address_bytes)
        if address_type == TYPE_INVALID:
            raise InvalidConfluxHexAddress
        if verbose:
            parts.append(f"{TYPE}.{address_type}")
        parts.append(payload + checksum)
        address = DELIMITER.join(parts)
        if verbose:
            return address.upper()
        return address

    @classmethod
    def _decode_network_id(cls, base32_address: str) -> int:
        base32_address = base32_address.lower()
        parts = base32_address.split(DELIMITER)
        return cls._network_prefix_to_id(parts[0])

    @classmethod
    def _decode_hex_address(cls, base32_address: str) -> HexAddress:
        base32_address = base32_address.lower()
        parts = base32_address.split(DELIMITER)
        return cls._payload_to_hex(parts[-1])

    @classmethod
    def _payload_to_hex(cls, payload: str) -> HexAddress:
        address_buf = base32.decode(payload)
        hex_buf = address_buf[1:21]
        return HEX_PREFIX + hex_buf.hex() # type: ignore

    # @classmethod
    # def _decode_address_type(cls, base32_address: str) -> AddressType:
    #     base32_address = base32_address.lower()
    #     hex_address = cls._decode_hex_address(base32_address)
    #     address_type = cls._detect_address_type(hex_address_bytes(hex_address))
    #     return address_type
    
    @classmethod
    def decode(cls, base32_address: str) -> Base32AddressParts:
        """
        Decode a base32 address string and get its hex_address, address_type, and network_id

        :param str base32_address: address encoded in base32 format
        :raises InvalidBase32Address:
        :return Base32AddressParts: a dict object with field "hex_address", "address_type" and "network_id"
        
        :examples:
        
        >>> Base32Address.decode("cfxtest:aatp533cg7d0agbd87kz48nj1mpnkca8be1rz695j4")
        {'network_id': 1, 'hex_address': '0x1ecde7223747601823f7535d7968ba98b4881e09', 'address_type': 'user'}
        """        
        try:
            if not isinstance(base32_address, str):
                raise InvalidBase32Address(f"Receives an argument of type {type(base32_address)}, expected a string")
            if not any([
                str(base32_address) == base32_address.upper(), 
                str(base32_address) == base32_address.lower(), 
            ]):
                raise InvalidBase32Address("Base32 address is supposed to be composed of all uppercase or lower case, "
                                        f"Receives {base32_address}")
            
            base32_address = base32_address.lower()

            splits = base32_address.split(DELIMITER)
            if len(splits) != 2 and len(splits) != 3:
                raise InvalidBase32Address(
                    "Address needs to be encode in Base32 format, such as cfx:aaejuaaaaaaaaaaaaaaaaaaaaaaaaaaaajrwuc9jnb\n"
                    "Received: {}".format(base32_address))

            # if exception occurs, outer try-except will handle
            address_parts = cls._decode(base32_address)

            # check address type
            address_type = address_parts["address_type"]
            if address_type == TYPE_INVALID:
                raise InvalidBase32Address(f"Invalid address type: the hex address of the provided address is {address_parts['hex_address']}, "
                                        "while valid conflux address is supposed to start with 0x0, 0x1 or 0x8")
            
            """
            cip-37 #Decoding
            6.Verify optional fields:

            If the optional fields contain type.*: Verify the address-type according to the specification above.
            Unknown options (options other than type.*) should be ignored.
            
            which means
            if address field is not expected, reject if the address field is a known one, else ignore
            e.g.
            valid CFX:TYPE.USER:AATP533CG7D0AGBD87KZ48NJ1MPNKCA8BE7GGP3VPU
            invalid CFX:TYPE.NULL:AATP533CG7D0AGBD87KZ48NJ1MPNKCA8BE7GGP3VPU
            ignore(but valid) CFX:GOD:AATP533CG7D0AGBD87KZ48NJ1MPNKCA8BE7GGP3VPU
            """
            if len(splits) == 3 and splits[1] != f"{TYPE}.{address_type}":
                address_field = splits[1]
                if address_field.startswith(f"{TYPE}.") and address_field[5:] in get_args(AddressType):
                    raise InvalidBase32Address(f"Invalid address type field: the address type field does not match expected, "
                                            f"expected {TYPE}.{address_type} but receives {address_field}, which is a known address type")
            
            # check checksum
            hex_address = address_parts["hex_address"]
            address_bytes = hex_address_bytes(hex_address)
            payload = base32.encode(VERSION_BYTE + address_bytes)
            checksum = cls._create_checksum(splits[0], payload)
            if checksum != base32_address[-8:]:
                raise InvalidBase32Address(f"Invalid Base32 address: checksum verification failed")
            return address_parts
        except Exception as e:
            if isinstance(e, InvalidBase32Address):
                raise e
            else:
                # unexpected error
                raise InvalidBase32Address(
                    "Address needs to be a Base32 format string, such as cfx:aaejuaaaaaaaaaaaaaaaaaaaaaaaaaaaajrwuc9jnb\n"
                    f"Received argument {base32_address} of type {type(base32_address)}")
        
    @classmethod
    def _decode(cls, base32_address: str) -> Base32AddressParts:
        """ 
        do not validate unless necessary, used if validity is known
        """
        parts = base32_address.split(DELIMITER)
        network_id = cls._network_prefix_to_id(parts[0])
        hex_address = cls._payload_to_hex(parts[-1])
        address_type = cls._detect_address_type(hex_address_bytes(hex_address))
        return {
            "network_id": network_id,
            "hex_address": hex_address,
            "address_type": address_type
        }

    @classmethod
    def is_valid_base32(cls, value: Any) -> bool:
        """
        Whether a value is a valid string-typed base32 address
        
        :return bool: True if valid, else False
        """        
        try:
            cls.decode(value)
            return True
        except:
            return False

    @classmethod
    def validate(cls, value: Any) -> Literal[True]:
        """
        validate if a value is a valid string-typed base32_address, raises an exception if not

        :param str base32_address: address to validate 
        :raises InvalidBase32Address: raises an exception if the address is not a valid base32 address
        :return Literal[True]: returns True only if address is valid
        """
        # an exception will be raised if decode failure
        cls.decode(value)
        return True
  
    @classmethod
    def equals(cls, address1: str, address2: str) -> bool:
        """
        check if two addresses share same hex_address and network_id

        :param str address1: base32 address to compare
        :param str address2: base32 address to compare
        :raises InvalidBase32Address: either address is not a valid base32 address
        :return bool: whether two addresses share same hex_address and network_id
        """
        return Base32Address(address1) == address2
    
    @classmethod
    def zero_address(cls, network_id: int, verbose: bool = False) -> "Base32Address":    
        """
        Get zero address of the target network.

        :param int network_id: target network_id
        :param bool verbose: whether the zero address is presented in verbose mode, defaults to False
        :return Base32Address: base32 format zero address of the target network 
        :examples:
        
        >>> Base32Address.zero_address(network_id=1)
        'cfxtest:aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa6f0vrcsw'
        """        
        return cls.encode("0x0000000000000000000000000000000000000000", network_id, verbose)
    
    @classmethod
    def shorten_base32_address(cls, base32_address: str, compressed=False) -> str:
        """
        returns the abbreviation of the address

        :param str base32_address: address to shorten
        :raises InvalidBase32Address: raised if address is invalid
        :param bool compressed: whether the abbreviation will be presented in compressed form,\
            which only affects mainnet addresses, defaults to False
        :return str: the abbreviation string
        
        >>> Base32Address.shorten_base32_address("cfx:aatp533cg7d0agbd87kz48nj1mpnkca8be7ggp3vpu")
        'cfx:aat...7ggp3vpu'
        >>> Base32Address.shorten_base32_address("cfx:aatp533cg7d0agbd87kz48nj1mpnkca8be7ggp3vpu", compressed=True)
        'cfx:aat...3vpu'
        """        
        cls.validate(base32_address)
        return cls._shorten_base32_address(base32_address, compressed)
    
    @classmethod
    def _shorten_base32_address(cls, base32_address: str, compressed=False) -> str:
        lowcase = base32_address.lower()
        splits = lowcase.split(DELIMITER)
        prefix = splits[0]
        payload = splits[-1]
        if splits[0] != MAINNET_PREFIX or compressed:
            return f"{prefix}{DELIMITER}{payload[:3]}...{payload[-4:]}"
        else:
            return f"{prefix}{DELIMITER}{payload[:3]}...{payload[-8:]}"
    
    @classmethod
    def calculate_mapped_evm_space_address(cls, base32_address: str) -> HexAddress:
        """
        calculate the address of mapped account for EVM space as defined in https://github.com/Conflux-Chain/CIPs/blob/master/CIPs/cip-90.md#mapped-account

        :raises InvalidBase32Address: raised when address is not a valid base32 address
        :examples:
        
        >>> Base32Address.calculate_mapped_evm_space_address("CFXTEST:TYPE.USER:AATP533CG7D0AGBD87KZ48NJ1MPNKCA8BE1RZ695J4")
        '0x349f086998cF4a0C5a00b853a0E93239D81A97f6'
        """        
        hex_address = cls.decode(base32_address)["hex_address"]
        return cls._mapped_evm_address_from_hex(hex_address)

    @classmethod
    def _mapped_evm_address_from_hex(cls, hex_address: HexAddress) -> HexAddress:
        # do not check hex_address validity here
        mapped_hash = keccak(HexBytes(hex_address)).hex()
        mapped_evm_space_address = to_checksum_address(HEX_PREFIX + mapped_hash[-40:])
        return cast(HexAddress, mapped_evm_space_address)

    @classmethod
    def _encode_network_prefix(cls, network_id: int) -> NetworkPrefix:
        if network_id == MAINNET_NETWORK_ID:
            return MAINNET_PREFIX
        elif network_id == TESTNET_NETWORK_ID:
            return TESTNET_PREFIX
        else:
            return COMMON_NET_PREFIX + str(network_id)

    @classmethod
    def _network_prefix_to_id(cls, network_prefix: NetworkPrefix) -> int:
        network_prefix = network_prefix.lower()
        if network_prefix == MAINNET_PREFIX:
            return MAINNET_NETWORK_ID
        elif network_prefix == TESTNET_PREFIX:
            return TESTNET_NETWORK_ID
        else:
            if not network_prefix.startswith(COMMON_NET_PREFIX):
                raise InvalidBase32Address(f"The network prefix {network_prefix} is invalid")
            try:
                return int(network_prefix[3:])
            except:
                raise InvalidBase32Address(f"The network prefix {network_prefix} is invalid")

    @classmethod
    def _create_checksum(cls, prefix, payload) -> str:
        """
        create checksum from prefix and payload
        :param prefix: network prefix (string)
        :param payload: bytes
        :return: string
        """
        prefix = cls._prefix_to_words(prefix)
        delimiter = VERSION_BYTE
        payload = base32.decode_to_words(payload)
        template = CHECKSUM_TEMPLATE
        mod = cls._poly_mod(prefix + delimiter + payload + template)
        return base32.encode(cls._checksum_to_bytes(mod))

    @classmethod
    def _detect_address_type(cls, hex_address_buf) -> AddressType:
        if hex_address_buf == bytes(20):
            return TYPE_NULL
        first_byte = hex_address_buf[0] & 0xf0
        if first_byte == 0x00:
            return TYPE_BUILTIN
        elif first_byte == 0x10:
            return TYPE_USER
        elif first_byte == 0x80:
            return TYPE_CONTRACT
        else:
            return TYPE_INVALID

    @classmethod
    def _prefix_to_words(cls, prefix) -> bytearray:
        words = bytearray()
        for v in bytes(prefix, 'ascii'):
            words.append(v & 0x1f)
        return words

    @classmethod
    def _checksum_to_bytes(cls, data) -> bytearray:
        result = bytearray(0)
        result.append((data >> 32) & 0xff)
        result.append((data >> 24) & 0xff)
        result.append((data >> 16) & 0xff)
        result.append((data >> 8) & 0xff)
        result.append((data) & 0xff)
        return result

    @classmethod
    def _poly_mod(cls, v: Union[bytes, bytearray]) -> int:
        """
        :param v: bytes
        :return: int64
        """
        assert type(v) == bytes or type(v) == bytearray
        c = 1
        for d in v:
            c0 = c >> 35
            c = ((c & 0x07ffffffff) << 5) ^ d
            if c0 & 0x01:
                c ^= 0x98f2bc8e61
            if c0 & 0x02:
                c ^= 0x79b76d99e2
            if c0 & 0x04:
                c ^= 0xf33e5fb3c4
            if c0 & 0x08:
                c ^= 0xae2eabe2a8
            if c0 & 0x10:
                c ^= 0x1e4f43e470

        return c ^ 1
